// Control of Chaos Pendulum
// july 2016 jan@breem.nl, janbee@hack42.nl

#include <arduino.h>
#include "DiagNano.h"

// Target: Arduino NANO 328
// Pin use:
// A0, PC0, Opto sensor A
// A1, PC1, Opto sensor B
// A2, PC2, Select motor phase W
// A3, PC3, Select motor phase V
// A4, PC4, Select motor phase U 
// A5, PC5, Led indicating opto B
// D13, PB5, Led indicating opto A

// Resource use:
// Timer 0 : system as usual
// Timer 1 : 16-bit timer calls an interrupt routine which does most of the work
// USART 0 : Programming, serial communication 

// When the pendulum passes through the opto slots the time to get from A to B
// is measured in counts. This number of counts is multiplied by
#define Stretch1 4
#define Stretch2 3
// to set the time for the kickpulses

// The maximum time of no movement detected
// Exceding this time will start the start-up procedure
// The time is measured in interrupt counts, so 4000 counts == 1 second 
// Set to 8 seconds:
#define MaxStallTime 32000

// Maximum number of complete loopings before suspend
// This prevents the pendulum from acquiring to much energy
#define MaxLoopings 4

int ReportInterval, ReportPassTicks;
bool  Hold; // Diagnostics. Everything goes normal, except the motor is not controlled 
String Message;
bool SendMessage;

//********************************************************************************************************
void PhaseU(bool On) 
{
  if(On) PORTC |= 0x10; else PORTC &= ~0x10; // PORT C bit 4 drives motor U phase
}

//********************************************************************************************************
void PhaseV(bool On)
{
  if(On) PORTC |= 0x08; else PORTC &= ~0x08; // PORT C bit 3 drives motor V phase
}

// ********************************************************************************************************
void PhaseW(bool On)
{
  if(On) PORTC |= 0x04; else PORTC &= ~0x04; // PORT C bit 2 drives motor W phase
}

// ********************************************************************************************************
ISR(TIMER1_COMPA_vect) // called at 4 kHz, execution takes approx 5us
{
  static long Counter, StallCounter, StallCounterA, StallCounterB; 
  static int StartKickTicks, Kick1Ticks, Kick2Ticks, SameDirSeen;                  
  static byte Try;
  static bool A, B, Dir, Suspend, DoStartProcedure, StartProcedureBusy;
  static enum {Idle, SeenA, SeenB, Kick1, Kick2} State;
   
  DIAGpin15H // indicate In Service Time on pin 15 (D12): around 5 usec
  
  // Find out which opto interruptor is active, 
  // Detect stall conditions and if found, execute the startup procedure 
  if(!DoStartProcedure)
  {
    A = (PINC & 0x01) == 0x01; // shutter present = high
    if(A) 
    {
      PORTB &= ~0x20; // low = led on PORT B bit 5 on
      StallCounter = 0;  // clear other stallcounters
      StallCounterB = 0;
      // test for A activated to long 
      if(++StallCounterA > MaxStallTime) DoStartProcedure = true;
    } 
    else PORTB |= 0x20; // led on PORT B bit 5 off
    
    B = (PINC & 0x02) == 0x02; // shutter present = high
    if(B)
    {
      PORTC &= ~0x20; // low = led on PORT C bit 5 on 
      StallCounter = 0;  // clear other stallcounters
      StallCounterA = 0;
      // test for B activated to long
      if(++StallCounterB > MaxStallTime) DoStartProcedure = true;
    } 
    else PORTC |= 0x20; // low = led on PORT C bit 5 on
    
    // test for to long without activity
    if(++StallCounter > MaxStallTime) DoStartProcedure = true;
  }  
  else // DoStartProcedure
  {
    if(!StartProcedureBusy)
    {
      StartKickTicks = 1000;
      if(!Hold)  // if we are on hold don't touch the motor
      {
        // The pendulum might be stalled in line with one of the phases
        // so try them all until one works.
        if(Try == 0) PhaseU(true);
        if(Try == 1) PhaseV(true);
        if(Try == 2) PhaseW(true);
        if(++Try > 2) Try=0;
      }
      StartProcedureBusy = true;
    }
    else
    {
      if(--StartKickTicks < 0) // Check end of startpulse
      { 
        if(!Hold) // if we are on hold don't touch the motor 
        {
          PhaseU(false);
          PhaseV(false);
          PhaseW(false);
        }
        StallCounter = 0;
        StallCounterA = 0;
        StallCounterB = 0; 
        DoStartProcedure = false;
        StartProcedureBusy = false;
      }
    }
  }
   
  Counter++;
  
  // This is the engine which kicks the pendulum when 
  // it passes through the opto interruptors.
  // Suspend kicking if found to many loopings in the same direction
  switch (State)
  {
    case Idle:
      if(A) // seen A first?
      {
        ReportInterval = Counter; // time since last shutter pass
        Counter = 0;
        if(SameDirSeen < 0) SameDirSeen = 0; // A after several B's
        Suspend = (++SameDirSeen > MaxLoopings); // skip kicking
        State = SeenA;
      }
      if(B) // seen B first?
      {
        ReportInterval = Counter; // time since last shutter pass
        Counter = 0;
        if(SameDirSeen > 0) SameDirSeen = 0; // B after several A's
        Suspend = (--SameDirSeen < -MaxLoopings); // skip kicking
        State = SeenB;
      }
    break;
    
    case SeenA: // so wait for B
      if(B) // start Kick 1
      {
        if(!Hold && !Suspend) PhaseU(true); 
        ReportPassTicks = Counter;
        Kick1Ticks = Counter * Stretch1; 
        Kick2Ticks = Counter * Stretch2; 
        //Message = "AB "; 
        SendMessage = true;
        Dir = true;
        State = Kick1;
      }  
    break;
    
    case SeenB:  // so wait for A
      if(A) // start Kick 1
      {
        if(!Hold && !Suspend) PhaseW(true);
        ReportPassTicks = -Counter;
        Kick1Ticks = Counter * Stretch1; 
        Kick2Ticks = Counter * Stretch2;
        //Message = "BA "; 
        SendMessage = true;
        Dir = false;
        State = Kick1;
      }  
    break;
      
    case Kick1: // the first kick after passing the opto interruptors
      if(--Kick1Ticks < 0) // wait for Kick 1 to end, then start Kick 2
      {
        if(!Hold && !Suspend) // otherwise skip activating motor windings
          if(Dir) // we had first kick U, now comes kick W
          {
            PhaseU(false);
            PhaseW(true);          
          }
          else // we had first W, now comes U
          {
            PhaseW(false);
            PhaseU(true);
          }
        //Message = "K1K2 "; SendMessage = true;
        State= Kick2;
      }
    break;
    
    case Kick2: 
      // the second kick should be timed such that the pendulum is 
      // out of the phase range of the first kick 
      if(--Kick2Ticks < 0) // wait for end of Kick 2 
      {
        if(!Hold) // if we are on hold don't touch the motor
        {
          PhaseU(false);
          PhaseV(false);
          PhaseW(false);
        }
        //Message = "end K2 "; SendMessage = true;
        State = Idle;
      }   
    break;
  }
  DIAGpin15L  // end of In Service flag
}
 
// ********************************************************************************************************
void setup()
{
  Serial.begin(19200);
  Serial.println("Hallo");
  
  // Configure 16 bit Timer 1 for 5 kHz rate 
  TCCR1A = 0x00;  // Waveform mode 12, no PWM outputs, Reload = OCR1
  TCCR1B = 0x0B;  // Internal clock div 64 gives 250 kHz 
  TIMSK1 = 0x22;  // Intr on compare A match
  OCR1A = 62;     // approx 4 kHz 
  
  DDRC = 0x3C;    // bit 0,1: inputs photointerruptors, 2,3,4: out motor UVW, 5: out interruptor led 
  DDRB = 0x20;    // PB5 out interruptor led
  PORTC = 0x03;   // Enable weak pullups for the photo interruptors
  
  PhaseU(false);  // to be sure
  PhaseV(false);         
  PhaseW(false);
  
  UseDIAGpin15  // for diagnosis with e.g. oscilloscope
  UseDIAGpin13
  UseDIAGpin12
}

// ********************************************************************************************************
void loop()
{
  if(Serial.available() > 0) 
  {
    // Single character commands can be given from the monitor in the Arduino IDE or any other terminal emulator on 19200 baud
    // u, v, and w engage a single motor phase. Note that the rest position of the pendulum must be aligned with the v phase 
    // so activate the v phase while mounting the pendulum in its rest position. 
    int Command = Serial.read();
    if(Command == 117) {PhaseU(true); PhaseV(false); PhaseW(false); Hold = true;} // u
    if(Command == 118) {PhaseU(false); PhaseV(true); PhaseW(false); Hold = true;} // v
    if(Command == 119) {PhaseU(false); PhaseV(false); PhaseW(true); Hold = true;} // w
    if(Command == 105) {PhaseU(false); PhaseV(false); PhaseW(false); Hold = true;} // i for motor idle
    if(Command == 99)  {PhaseU(false); PhaseV(false); PhaseW(false); Hold = false;} // c continue to be chaotic
  }  
  
  if(SendMessage) 
  {
    // Serial.print(Message); // diagnostics
    // The following message format may be directly recorded as a comma separated file
    // which can be imported to e.g. excel to produce graphical presentations 
    Serial.print(ReportPassTicks);
    Serial.print(", ");
    Serial.println(ReportInterval); 
    SendMessage = false;
  }
}

